# : Hnsm machine stack N Hohensee ;  ( Hnsm)

# ksh emulation of 4-stack CPU. About 60 typeable ASCII opcodes.
# needs printf in Bash or in $PATH. Helpers use od, grep... 
# 2021 March; most opcodes kinda work
# This program is public domain.
# Rick presidentbyamendment Hohensee.

. ASC.sh

# To get the feel of a semi-typeable machine language, I want Hnsm code 
# to be actually binary, and sh just understands C text strings. Binary 
# zero is therefor impossible. Gotta have zero. SO, we swap 'z' and 
# zero in the ASCII routine. A byte value of 122 is therefor impossible 
# in literals. Otherwise, 'z' is a valid opcode. If you really need a 
# 122 byte in a literal... figure something out. Other memory will be 
# sh integers, which are apparently strings. 0 works. I doubt if 
# lacking 122 will be a problem, but be aware that z in a literal is a 
# zero.


pc=0 rsp=0x1000 fsp=0x1000 cl=0x100 asp=0x1000 as=0
csp=0x2000 cs=0 lp=0x10 macro=0

fs[$fsp]=0 fs[$fsp-1]=0 fs[$fsp-2]=0 fs[$fsp-3]=0 fs[$fsp-4]=0
fs[$fsp-5]=0 fs[$fsp-6]=0 fs[$fsp-7]=0 fs[$fsp-8]=0 fs[$fsp-9]=0
fs[$fsp-10]=0

S=0 Z=0 C=0 V=0
		# settable switches for addressing modes;
		#  post-pre, naked/writeback, incr/decr
PO=0 WB=0 DN=0

		#  pointer terminus flag
P=0

boole () (
for i in $*; do
	let j=i\&1
	echo -n $j" "
done
)


hex () {
	printf "%08x  " $1
}


hexdual () {
	printf "%04x " $1
}


ifhex () {
if test -v $1 ; then
hex 0
else
hex $1
fi
}


ifhexdual () {
if test -v $1 ; then
hexdual 0
else
hexdual $1
fi
}



PIC () {

echo "
"${code:$pc:60}
echo "^
PC	  Z S C V   P PO WB DN "
ifhex $pc
boole $Z
boole $S
boole $C
boole $V
echo -n "  "
boole $P
boole $PO
echo -n " "
boole $WB
echo -n " "
boole $DN
echo " loops
return    data      address   lim  cnt   coproc"
ifhex ${rs[$(($rsp))]}
ifhex ${fs[$fsp]}
ifhex ${as[$asp]}
ifhexdual ${lim[$lp]}
ifhexdual  ${lc[$lp]}
echo -n " "
ifhex ${cs[$csp]}
echo
ifhex ${rs[$(($rsp-1))]}
ifhex ${fs[$(($fsp-1))]}
ifhex ${as[$asp-1]}
ifhexdual ${lim[$lp-1]}
ifhexdual  ${lc[$lp-1]}
echo -n " "
ifhex ${cs[$(($csp-1))]}
echo
ifhex ${rs[$(($rsp-2))]}
ifhex ${fs[$(($fsp-2))]}
ifhex ${as[$asp-2]}
ifhexdual ${lim[$lp-2]}
ifhexdual  ${lc[$lp-2]}
echo -n " "
ifhex ${cs[$csp-2]}
echo
ifhex ${rs[$(($rsp-3))]}
ifhex ${fs[$(($fsp-3))]}
ifhex ${as[$asp-3]}
ifhexdual ${lim[$lp-3]}
ifhexdual  ${lc[$lp-3]}
echo -n " "
ifhex ${cs[$(($csp-3))]}
echo
ifhex ${rs[$(($rsp-4))]}
ifhex ${fs[$(($fsp-4))]}
ifhex ${as[$asp-4]}
ifhexdual ${lim[$lp-4]}
ifhexdual  ${lc[$lp-4]}
echo -n " "
ifhex ${cs[$(($csp-4))]}
echo
ifhex ${rs[$(($rsp-5))]}
ifhex ${fs[$(($fsp-5))]}
ifhex ${as[$asp-5]}
ifhexdual ${lim[$lp-5]}
ifhexdual  ${lc[$lp-5]}
echo -n " "
ifhex ${cs[$csp-5]}
echo
ifhex ${rs[$(($rsp-6))]}
ifhex ${fs[$(($fsp-6))]}
ifhex ${as[$asp-6]}
ifhexdual ${lim[$lp-6]}
ifhexdual  ${lc[$lp-6]}
echo -n " "
ifhex ${cs[$csp-6]}
echo
}


NEXT ()
{
    wv=${code:$pc:1}
    let pc=$pc+1
}


DROP ()
{
    let fsp-=1
}


BUMP ()
{
    let fsp+=1
}

LIT2 () 
{
	NEXT
	a=`ASCII $wv`
	NEXT
	b=`ASCII $wv`
	let e=b\<\<8\|a
	echo $e
}


LIT4 ()
{
NEXT
a=`ASCII $wv`
NEXT
b=`ASCII $wv`
NEXT
c=`ASCII $wv`
NEXT
d=`ASCII $wv`
let e=$d\<\<24\|$c\<\<16\|$b\<\<8\|$a
echo $e
}


ZEROTEST ()
{
    if test ${fs[$fsp]} -eq 0; then
        let Z=1
    else
        let Z=0
    fi
}


SIGNTEST ()
{
    if test $((${fs[$fsp]}&0x80000000)) -eq 0; then
        let S=0
    else
        let S=1
    fi
}




POINTERFLAG ()
{
	if  test ! ${as[$asp]} ; then
	echo "set some pointers"
	return 1
	fi
    if test ${as[$asp]} = ${as[$asp-1]}  -o ${as[$asp]} = "0" ; then
        let P=0
    else
        let P=1
    fi
}


LOOPTEST ()
{
    if test ${lim[$lp]} -eq ${lc[$lp]}; then
        let lt=0
    else
        let lt=1
    fi
}


HILOMUL ()
{
    let a=0 b=0 c=0 d=0
    let a=$1\&0xffff
    let b=($1\&0xffffffff)\>\>16
    let c=($2\&0xffffffff)\>\>16
    let d=$2\&0xffff
    let e=a*d
    let f=a*c
    let g=b*d
    let h=b*c
    let mid=g+f
    let lomid=(mid\<\<16)\&0xffffffff
    let himid=mid\>\>16
    let lo=lomid+e
    let hi=himid+h
}


ATOM ()
{
    set -f
    case $wv in
        " ")
            echo "space		\"nop\""
        ;;
        "!")
            echo "!		\"store\""
                case $PO$WB$DN in
                    000)
                        let mem[${as[$asp]}]=${fs[$fsp]}
                    ;;
                    001)
                        let mem[${as[$asp]}]=${fs[$fsp]}
                    ;;
                    010)
                        let mem[${as[$asp]}]=${fs[$fsp]}
                        let as[$asp]+=1
                    ;;
                    011)
                        let mem[${as[$asp]}]=${fs[$fsp]}
                        let as[$asp]-=1
                    ;;
                    100)
                        let mem[${as[$asp]}+1]=${fs[$fsp]}
                    ;;
                    101)
                        let mem[${as[$asp]}-1]=${fs[$fsp]}
                    ;;
                    110)
                        let as[$asp]+=1
                        let mem[${as[$asp]}]=${fs[$fsp]}

                    ;;
                    111)
                        let as[$asp]=${as[$asp]}-1
                        let mem[${as[$asp]}]=${fs[$fsp]}
                    ;;
                esac
            DROP
        ;;

        "\"")
            echo "\"		\"lit\""
            BUMP
            fs[$fsp]=`LIT4`
            let pc+=4
            ZEROTEST
            SIGNTEST
        ;;

        "#")
            echo "#		\"comment\""
            while test "$wv" != "#"; do
                NEXT
            done
        ;;
        "$")
            echo "$		\"syscall\""
            echo $pc $wv
            while test "$wv" != "$"; do
                NEXT
		syscall+=$wv
            done
        ;;
        "%")
         # A B C --- wide(B*C)>>A  Forth  */ with a downshift.
         # 2 instructions on 386. 1 on Hnsm.
            echo "%		\"scale\""
            a=${fs[$fsp]}
            let b=a\|31
	    let c=32-b
            DROP
            HILOMUL ${fs[$fsp]} ${fs[$fsp-1]}
            let d=lo\>\>b
	    let e=hi\<\<c
            DROP
	    let fs[$fsp]=d\|e
            ZEROTEST
            SIGNTEST
        ;;
        "&") echo "&		\"and\""
            let fs[$fsp-1]\&=${fs[$fsp]}
            DROP
            ZEROTEST
            SIGNTEST
        ;;
        "(")
            echo "(		\"for/until\""
	   let lp+=1
           lim[$lp]=`LIT2`
           let pc+=2
           lc[$lp]=`LIT2`
           let pc+=2
        ;;
        ")")
            echo -n ")		\"loop\""
            LOOPTEST
            if test $lt == "1"; then
                if test ${lim[$lp]} -eq 0; then
                    let lc[$lp]-=1\&0xffffffff
                    echo "for  "
                else
                    echo "until"
                    let lc[$lp]+=1\&0xffffffff
                fi
                let pc=\(pc-`LIT2`\)\&0xffffffff
            else
                let lp-=1
            fi
        ;;
        "*")
            echo "*		\"times\""
            HILOMUL ${fs[$fsp]} ${fs[$fsp-1]}
	    let fs[$fsp]=hi
	    let fs[$fsp-1]=lo
            ZEROTEST
            SIGNTEST
        ;;
        +)
            echo "+		\"plus\""
            let fs[$fsp-1]+=${fs[$fsp]}\&0xffffffff
            DROP
            ZEROTEST
            SIGNTEST
        ;;
        ,)
            echo ",		\"nop\""
        ;;
        "'")
            echo "\'		\"nor\""
            let fs[$fsp-1]\|=${fs[$fsp]}
            let fs[$fsp-1]=~${fs[$fsp-1]}\&0xffffffff
            DROP
            ZEROTEST
            SIGNTEST
        ;;
        -)
            echo "-		\"minus\""
            let fs[$fsp-1]=\(${fs[$fsp-1]}-${fs[$fsp]}\)\&0xffffffff
            DROP
            ZEROTEST
            SIGNTEST
        ;;
        .)
            echo ".		\"halt\""
	#	suspend -f ?
        ;;
        /)
            echo "/ 		\"divmod\""
            let a=${fs[$fsp-1]}
            let fs[$fsp-1]=${fs[$fsp-1]}/${fs[$fsp]}
            let fs[$fsp]=$a%${fs[$fsp]}
            ZEROTEST
            SIGNTEST
        ;;
       0)
            echo "0		\"zero\""
            BUMP
            let fs[$fsp]=0
            Z=1
            S=0
        ;;
        1)
            echo "1		\"increment\""
            let fs[$fsp]+=1
            ZEROTEST
            SIGNTEST
        ;;
        2)
            echo "2		\"times 2\""
	    if test $((${fs[$fsp]}&0X80000000))  -eq 0; then
	      let C=0
	    else
	       let C=1
	    fi
            let fs[$fsp]=${fs[$fsp]}\<\<1\&0xffffffff
            ZEROTEST
	    SIGNTEST
        ;;
        3)
            echo "3		\"div 3\""
            let a=0
            a=${fs[$fsp]}
            BUMP
            let fs[$fsp]=$a%3
            let fs[$fsp-1]=$a/3
            ZEROTEST
            SIGNTEST
        ;;
        4) echo "4	nop"
        ;;
        5)
            echo "5		\"div 5\""
           let a=0
            a=${fs[$fsp]}
            BUMP
            let fs[$fsp]=$a%5
            let fs[$fsp-1]=$a/5
            ZEROTEST
            SIGNTEST
        ;;
        6)    echo "6	nop"
        ;;
        7)  echo "7		\"jarble\""
        #   A  B  --- swap bit B in A with bit 0 in A
        # so does nothing if they're the same.
	DROP
	let j=fs[$fsp+1]\&31
	let bn=1\<\<j
	let b0=fs[$fsp]\&1
	let mask=b0\|bn
	let a=\!bn
	if test $a -ne $b0 ; then
	     let fs[$fsp]=v^mask
	fi
        ;;
        8)  echo nop
        ;;
        9) echo nop
        ;;
        ":")
            echo ":		\"call\""
	    let rsp+=1
            let rs[$rsp]=pc+4
            pc=`LIT4`
        ;;
        ";")
            echo ";		\"return\""
            let pc=rs[$rsp]
            let rsp-=1
        ;;
        "<")
            echo "<		\"min\""
            if test ${fs[$fsp]} -le ${fs[$fsp-1]}; then
                let fs[$fsp-1]=${fs[$fsp]}
            fi
            DROP
            ZEROTEST
            SIGNTEST
        ;;
        "=")
            echo "=		\"nexor\""
            let fs[$fsp-1]\^=~${fs[$fsp]}\&0xffffffff
            DROP
            ZEROTEST
            SIGNTEST
        ;;
        ">")
            echo ">		\"max\""
            if test ${fs[$fsp]} -gt ${fs[$fsp-1]}; then
                let fs[$fsp-1]=${fs[$fsp]}
            fi
            DROP
            ZEROTEST
            SIGNTEST
        ;;
        "?")
            echo "?		\"if\""
            NEXT
            condition=$wv
            branch=`LIT2`
            let pc+=2
            if test $branch -gt $((0x7fff)) ; then
               let branch=$((-branch))
            fi

            case $condition in
                Z)
                   if test $Z = "1"; then
                        let pc+=$branch
                    fi
                ;;
                z)
                    if test $Z = "0"; then
                        let pc+=$branch
                    fi
                ;;
                S|N)
                    if test $S = "1"; then
                        let pc+=$branch
                    fi
                ;;
                s|n)
                    if test $S = "0"; then
                        let pc+=$branch
                    fi
                ;;
                C)
                    if test $C = "1"; then
                        let pc+=$branch
                    fi
                ;;
                c)
                    if test $C = "0"; then
                        let pc+=$branch
                    fi
                ;;
                O|V)
                    if test $V = "1"; then
                        let pc+=$branch
                    fi
                ;;
                o|v)
                    if test $Z = "0"; then
                        let pc+=$branch
                    fi
                ;;
                P)
                    if test $P = "1"; then
                        let pc+=$branch
                    fi
                ;;
                p)
                    if test $P = "0"; then
                        let pc+=$branch
                    fi
                ;;
                G|g)
                    if test $C = "1" -a $Z = "0"; then
                        let pc+=$branch
                    fi
                ;;
                L|l)
                    if test $C = "0" -o $Z = "1"; then
                        let pc+=$branch
                    fi
                ;;
                m)
                    if test $N = $V; then
                        let pc+=$branch
                    fi
                ;;
                n)
                    if test $N != $V; then
                        let pc+=$branch
                    fi
                ;;
                q) # skip p and o for use above
                    if test $Z = "0" -a $N = $V; then
                        let pc+=$branch
                    fi
                ;;
                r)
                    if test $Z = "1" -o $N != $V; then
                        let pc+=$branch
                    fi
                ;;
                A)
                        let pc+=$branch
                ;;
            esac
        ;;
        @)
            echo "@		\"fetch\""
            BUMP
            #     pre/wb/dn
            case $PO$WB$DN in
                000)
                    let fs[$fsp]=mem[${as[$asp]}]
                ;;
                001)
                    let fs[$fsp]=mem[${as[$asp]}]
                ;;
                010)
                    let fs[$fsp]=mem[${as[$asp]}+1]
                    let as[$asp]+=1
                ;;
                011)
                    let fs[$fsp]=mem[${as[$asp]}-1]
                    let as[$asp]-=1
                ;;
                100)
                    let fs[$fsp]=mem[${as[$asp]+1}]
                ;;
                101)
                    let fs[$fsp]=mem[${as[$asp]}-1]
                ;;
                110)
                    let as[$asp]+=1
                    let fs[$fsp]=mem[${as[$asp]}]
                ;;
                111)
                    let as[$asp]-=1
                    let fs[$fsp]=mem[${as[$asp]}]
                ;;
            esac
        ;;
        A)
            echo "A		\"A mode\""
            NEXT
            AMODE=$wv
            case $wv in
                0)
                    let PO=0 WB=0 DN=0
                ;;
                1)
                    let PO=0 WB=0 DN=1
                ;;
                2)
                    let PO=0 WB=1 DN=0
                ;;
                3)
                    let PO=0 WB=1 DN=1
                ;;
                4)
                    let PO=1 WB=0 DN=0
                ;;
                5)
                    let PO=1 WB=0 DN=1
                ;;
                6)
                    let PO=1 WB=1 DN=0
                ;;
                7)
                    let PO=1 WB=1 DN=1
                ;;
            esac
        ;;
        B)
            echo "B		\"revbytes\""
		let c=0
		let a=fs[$fsp]

		let b=a\&255
		let c=c\<\<8
		let c=c\|b
		let a=a\>\>8

		let b=a\&255
		let c=c\<\<8
		let c=c\|b
		let a=a\>\>8

		let b=a\&255
		let c=c\<\<8
		let c=c\|b
		let a=a\>\>8

		let b=a\&255
		let c=c\<\<8
		let c=c\|b
		let fs[$fsp]=$c
		ZEROTEST
                SIGNTEST

        ;;
        C)
            echo "C		\"copro\""
            while test "$wv" != "C"; do
                NEXT
            done
        ;;
        D)
            echo "D		\"dupe\""
            BUMP
            let fs[$fsp]=${fs[$fsp-1]}
        ;;
        E)  echo nop
        ;;
        F)           echo nop
        ;;
        G)
            echo "G		\"go to\""
            let pc=`LIT4`
        ;;
        H)    echo nop
        ;;
        I)    echo nop
        ;;
        J)
            echo "J		\"jump V\""
		let pc=${fs[$fsp]}
		DROP
        ;;
        K)
            echo nop
        ;;
        L)
            echo "L		\"macro if\""
            # ? in an M...M, 1 byte condition, 1 byte signed branch
            NEXT
            condition=$wv
            NEXT
            let branch=$wv
            if test $((128&$branch)) -eq 128 ; then
		let branch=$((0-$branch))   # signed byte
	    fi
            case $condition in
                Z)
                   if test $Z = "1"; then
                        let mpc+=$branch
                    fi
                ;;
                z)
                    if test $Z = "0"; then
                        let mpc+=$branch
                    fi
                ;;
                S|N)
                    if test $S = "1"; then
                        let mpc+=$branch
                    fi
                ;;
                s|n)
                    if test $S = "0"; then
                        let mpc+=$branch
                    fi
                ;;
                C)
                    if test $C = "1"; then
                        let mpc+=$branch
                    fi
                ;;
                c)
                    if test $C = "0"; then
                        let mpc+=$branch
                    fi
                ;;
                O|V)
                    if test $V = "1"; then
                        let mpc+=$branch
                    fi
                ;;
                o|v)
                    if test $Z = "0"; then
                        let mpc+=$branch
                    fi
                ;;
                P)
                    if test $P = "1"; then
                        let mpc+=$branch
                    fi
                ;;
                p)
                    if test $P = "0"; then
                        let mpc+=$branch
                    fi
                ;;
                G|g)
                    if test $C = "1" -a $Z = "0"; then
                        let mpc+=$branch
                    fi
                ;;
                L|l)
                    if test $C = "0" -o $Z = "1"; then
                        let mpc+=$branch
                    fi
                ;;
                m)
                    if test $N = $V; then
                        let mpc+=$branch
                    fi
                ;;
                n)
                    if test $N != $V; then
                        let mpc+=$branch
                    fi
                ;;
                q) # skip p and o for use above
                    if test $Z = "0" -a $N = $V; then
                        let mpc+=$branch
                    fi
                ;;
                r)
                    if test $Z = "1" -o $N != $V; then
                        let mpc+=$branch
                    fi
                ;;
                A)
                        let mpc+=$branch
                ;;
            esac
        ;;
        M)
            echo "M		\"macro\""
            NEXT
            a=`ASCII $wv`
            while test "$wv" != "M"; do
                molecule[$a]+=$wv
                NEXT
            done
        ;;
        N)  echo "N		\"nip\""
            DROP
            let fs[$fsp]=${fs[$fsp+1]}
            ZEROTEST
            SIGNTEST
        ;;
        O)
            echo "O		\"over\""
            BUMP
            let fs[$fsp]=${fs[$fsp-2]}
            ZEROTEST
            SIGNTEST
        ;;
        P)  echo nop
        ;;
        Q)  echo nop
        ;;
        R)
            echo "R		\"ratchet\""
            let asp+=1
            let as[$asp]=${mem[${as[$asp-1]}]}
            POINTERFLAG
        ;;
        S)
           echo "S		\"swap\""
            let a=0
            let a=${fs[$fsp]}
            let fs[$fsp]=${fs[$fsp-1]}
            let fs[$fsp-1]=$a
            ZEROTEST
            SIGNTEST
        ;;
        T)    echo "T		\"true\""
	BUMP
	let Z=0
	let S=1
	let fs[$fsp]=0xffffffff
        ;;
        U)   echo nop
        ;;
        V)
            echo "V		\"value\""
            SIGNTEST
            if test $S = 1 ; then
               let   fs[$fsp]=~fs[$fsp]+1\&0xffffffff
            fi
            ZEROTEST
	    S=0
        ;;
        W)   echo "W		\"nop\""
        ;;
        X)
	    echo "X		\"extend dual\""
            if test $((0x8000&${fs[$fsp]})) -eq $((0x8000)) ; then
		let fs[$fsp]=fs[$fsp]\|0xffff0000
	    fi 
	    SIGNTEST
	    ZEROTEST
        ;;
        Y)
            echo "Y		\"yank\""
            let as[$asp]=mem[as[$asp]]
            POINTERFLAG
        ;;
        Z)  echo nop
        ;;
        "[")
            echo "[		\"upshift\""
           let  fs[$fsp-1]=fs[$fsp-1]\<\<fs[$fsp]\&0xffffffff
            DROP
            ZEROTEST
            SIGNTEST
        ;;
        "\\")
            echo "\\		\"ADC\""
            let ADCA=fs[$fsp]
            let ADCB=fs[$fsp-1]
            let ADCAL=ADCA\&0XFFFF
            let ADCBL=ADCB\&0XFFFF
            let ADCAH=ADCA\>\>16
            let ADCBH=ADCB\>\>16
            let ADCSL=(ADCAL+ADCBL+C)\&0xffff
            let ADCSLC=(ADCAL+ADCBL+C)\>\>16
            let ADCHS=(ADCAH+ADCBH+ADCSLC)\&0xffff
            let ADCCF=(ADCAH+ADCBH+ADCSLC)\>\>16
            let ADCSUM=ADCSL+\(ADCHS\<\<16\)
	    let fs[$fsp]=$ADCSUM
            ZEROTEST
            SIGNTEST
            # needs to propagate a carry if there is one
        ;;
        "]")
            echo "]		\"downshift\""
            let fs[$fsp-1]=fs[$fsp-1]\>\>fs[$fsp]
            DROP
            ZEROTEST
            SIGNTEST
        ;;
        "^")
            echo "^		\"exor\""
            let fs[$fsp-1]=${fs[$fsp]}^fs[$fsp-1]
            DROP
            ZEROTEST
            SIGNTEST
        ;;
        _)
            echo "_		\"sinum\""
            let fs[$fsp-1]=fs[$fsp]-fs[$fsp-1]\&0xffffffff
	    DROP
            ZEROTEST
            SIGNTEST
        ;;
        "\`")
            echo "\`		\"nand\""
            let let fs[$fsp-1]\&=${fs[$fsp]}
            DROP
	    let fs[$fsp]=~fs[$fsp]\&0xffffffff
            ZEROTEST
            SIGNTEST
        ;;
        a)
            echo "a		\"a-drop\""
            let asp+=-1
            POINTERFLAG
        ;;
        b)    echo "b		\"bitpop\""
		BUMP
	    let fs[$fsp]=0
	    let a=fs[$fsp-1]
	    while test $a -ne 0; do
	        let fs[$fsp]+=a\&1
		  let a/=2;
	    done
        ;;
        c)  echo nop
        ;;
        d)
           echo "d		\"drop\""
            DROP
            ZEROTEST
            SIGNTEST
        ;;
        e)    echo nop
        ;;
        f)  echo nop
        ;;
        g)  echo nop
        ;;
        h)  echo nop
        ;;
        i)
	    echo "i		\"eye\""
            # push current loop count to TOS
            BUMP
            let fs[$fsp]=lc[$lp]
            ZEROTEST
        ;;
        j)      echo "j		nop"
        ;;
        k)    echo "k		nop"
        ;;
        l)  echo nop
        ;;
        m)  echo nop
        ;;
        n)  echo nop
        ;;
        o)  echo nop
        ;;
        p)  echo nop
        ;;
        q)  echo nop
        ;;
        r)  echo nop
        ;;
        s)  echo "s		\"sub b\""
            let fs[$fsp-1]=\(${fs[$fsp-1]}-${fs[$fsp]}-$C\)\&0xffffffff
            DROP
            ZEROTEST
            SIGNTEST

        ;;
        t)
            echo "t		\"trans\""
	NEXT
	let b=wv\&15
	case $b in
		0)	let a=rsp		;;
		1)	let a=sp		;;
		2)	let a=rs[$rsp]
			let rsp-=1		;;
		3)	let a=fs[$fsp]
			let fsp-=1		;;
		4)	let a=as[$asp]
			let asp-=1		;;
		5)	let a=asp		;;
		6)	let a=ls[$lsp]
			let lsp-=1		;;
		7)	let a=lsp		;;
		8)	let a=cs[$csp]
			let csp-=1		;;
		9)	let a=csp		;;
		10)	let a=PO		;;
		11)	let a=WB		;;
		12)	let a=DN		;;
		13)	let unimplemented=1	;;
		14)	let unimplemented=1	;;
		15)	let a=pc	;;
	esac
	let b=wv\>\>4
	case $b in
		0)	let rsp=a		;;
		1)	BUMP
			let fs[$fsp]=a	;;
		2)	let rsp+=1
			let rs[$rsp]=a	;;
		3)	let fsp+=1
			let fs[$fsp]=a	;;
		4)	let asp+=1
			let as[$asp]=a	;;
		5)	let asp=a		;;
		6)	let to/lim\to/cnt	;;
		7)	let lsp+=1
			let ls[$lsp]=a	;;
		8)	LET csp+=1
			let cs[$csp]=a	;;
		9)	let csp=a		;;
		10)	let PO=a\&1	;;
		11)	let WB=a\&1	;;
		12)	let DN=a\&1	;;
		13)	let unimplemented=1	;;
		14)	let unimplemented=1	;;
		15)	let pc=a		;;
	esac
        ;;
        u)  echo nop
        ;;
        v)  echo nop
        ;;
        w)  echo nop
        ;;
        x)
            echo "x		\"extend byte\""
            if test $((128&${fs[$fsp]})) -eq 128 ; then
		let fs[$fsp]=${fs[$fsp]}\|0xffffff00
	    fi
	    SIGNTEST
	    ZEROTEST
        ;;
        y)  echo nop
        ;;
        z)  echo "z		\"mac z\""
        # branch if TOS !0
           if test $Z -eq 0 ; then
		 let mpc=0
	   fi
        ;;
        "{")   echo "{		\"roll\""
	let a=fs[$fsp-1]\<\<fs[$fsp]\&0xffffffff
	let b=fs[$fsp-1]\>\>\(32-fs[$fsp]\)
	let fs[$fsp-1]=a\|b
	DROP
	SIGNTEST
        ;;
        "|")
            echo "|		\"or\""
            let fs[$fsp-1]\|=${fs[$fsp]}
            DROP
            ZEROTEST
            SIGNTEST
        ;;
        "}")   echo "}		\"eggroll\""
	SIGNTEST
	let bit0=S
	let C=S
	let fs[$fsp]=fs[$fsp]\<\<1\|bit0\&0xffffffff

        ;;
        "~")
            echo "~		\"not\""
            let fs[$fsp]=~${fs[$fsp]}\&0xffffffff
            ZEROTEST
            SIGNTEST
        ;;
    esac
    set +f
}


IPL () { # initial load binary program file to $code
# Be sure file ends with a newline. TWICE in asm.sh does that once.
	code= COUNT=0
	REPLY=
	while read  ; do
		code=$code$REPLY
		let COUNT+=1
	done
	echo ${#code}   $COUNT
	echo "echo \$code if you're curious."
}


MEXT ()
{
    let mpc+=1
    wv=${mcode:$mpc:1}
}


MOLECULE () {
		mpc=0
		mcode=${molecule[$a]}
		echo ${molecule[$a]}
                while 
                MEXT
                test -n $wv -a $wv != M  ;do
			ATOM
		done
}


OPCODE ()
{
	a=`ASCII $1`
	if test ${molecule[$a]} ; then
             MOLECULE
	else
             ATOM
	fi
}


STEP () {
       NEXT ; OPCODE $wv ; PIC
}


STEPS () {
while read
do
	pc=0
	code=$REPLY
	aaaa=${#code}
	while test $pc -le $aaaa ; do
              OPCODE
              NEXT
	done
	PIC
done
}


SALT () { 
molecule[68]="kjnk\@\#\$\%\6\%\^\&\*\(\)\-\_+=jnkjnM"
molecule[65]="kj;;;;;;;;,,,,,,,,,,,M"
molecule[66]="kjnlxxsfflkjkjnM"

wv=1

rs[$rsp]=0
fs[$fsp]=0
as[$asp]=0
cs[$csp]=0
lim[$lp]=0
lc[$lp]=333
rs[$rsp-1]=2
fs[$fsp-1]=22
fs[$fsp-2]=22
fs[$fsp-3]=22
fs[$fsp-5]=22
fs[$fsp-4]=2
fs[$fsp-6]=2
fs[$fsp-7]=2
as[$asp-1]=222
cs[$csp-1]=2222
lim[$lp-1]=22222
lc[$lp-1]=0

pc=0

code=120812ffggzzxx92384MkdbcsHBEDEEMMMM


echo "code "$code
echo "pc "$pc	"    wv "$wv	"    mpc "$mpc
}


LIST () {
grep \"\"  vm.sh
}

# getting started:$ . vm.sh $ LIST $ SALT; PIC; STEP; STEP;...

